from plone import api
from Products.PlonePAS.Extensions.Install import activatePluginInterfaces
from zope.component.hooks import getSite
from .interfaces import IHasWorkspace
from .interfaces import IWorkspace
from .pas import get_workspace_groups_plugin
from .pas import add_group
import logging

logger = logging.getLogger('collective.workspace')


def setup_pas(context):
    if context.readDataFile('collective.workspace.txt') is None:
        return

    site = getSite()
    if 'workspace_groups' not in site.acl_users:
        site.acl_users.manage_addProduct[
            'collective.workspace'].addWorkspaceGroupManager(
            'workspace_groups', 'collective.workspace Groups',
            )
        activatePluginInterfaces(site, 'workspace_groups')


def migrate_groups(context):
    workspace_groups = get_workspace_groups_plugin()
    if hasattr(workspace_groups, '_groups'):
        # Already migrated
        return

    # Reinitialize groups plugin because its base class changed
    workspace_groups.__init__(workspace_groups.id, workspace_groups.title)

    # Store existing workspace group membership in workspace_groups
    catalog = api.portal.get_tool('portal_catalog')
    for b in catalog.unrestrictedSearchResults(
            object_provides=IHasWorkspace.__identifier__):
        logger.info('Migrating workspace groups for {}'.format(b.getPath()))
        workspace = IWorkspace(b._unrestrictedGetObject())
        for group_name in set(workspace.available_groups):
            group_id = '{}:{}'.format(group_name.encode('utf8'), b.UID)
            title = '{}: {}'.format(group_name.encode('utf8'), b.Title)
            add_group(group_id, title)
        for m in workspace:
            new_groups = m.groups & set(workspace.available_groups)
            m._update_groups(set(), new_groups)



# coding=utf-8
from collective.workspace.interfaces import IWorkspace
from collective.workspace.testing import COLLECTIVE_WORKSPACE_INTEGRATION_TESTING  # noqa: E501
from plone import api
from plone.app.testing import SITE_OWNER_NAME
from plone.testing import z2

import unittest


class TestWorkspace(unittest.TestCase):
    layer = COLLECTIVE_WORKSPACE_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        z2.login(self.app['acl_users'], SITE_OWNER_NAME)
        self.user1 = api.user.create(
            email='user1@example.com',
            username='user1',
            password='123'
        )
        self.workspace = api.content.create(
            container=self.portal,
            type='Workspace',
            id='a-workspace'
        )
        self.ws = IWorkspace(self.workspace)

    def _get_roles_in_workspace(self):
        pmt = api.portal.get_tool('portal_membership')
        member = pmt.getMemberById(self.user1.getId())
        return member.getRolesInContext(self.workspace)

    def test_adding_workspace_creates_groups(self):
        group = self.portal.portal_groups.getGroupById(
            'Admins:' + self.workspace.UID())
        self.assertIsNotNone(group)

    def test_renaming_workspace_updates_group_titles(self):
        self.workspace.setTitle('new title')
        from zope.lifecycleevent import modified
        modified(self.workspace)
        group = self.portal.portal_groups.getGroupById(
            'Admins:' + self.workspace.UID())
        self.assertEqual(group.getProperty('title'), 'Admins: new title')

    def test_removing_workspace_removes_groups(self):
        self.portal.manage_delObjects(['a-workspace'])
        group = self.portal.portal_groups.getGroupById(
            'Admins:' + self.workspace.UID())
        self.assertIsNone(group)

    def test_add_to_team(self):
        self.ws.add_to_team(
            user=self.user1.getId()
        )
        self.assertIn(self.user1.getId(), list(self.ws.members))

    def test_adding_team_member_updates_groups(self):
        self.ws.add_to_team(
            user=self.user1.getId(),
            groups=(u'Admins',),
            )
        self.assertIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Members:' + self.workspace.UID()),
        )
        self.assertIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Admins:' + self.workspace.UID()),
        )

    def test_updating_team_member_updates_groups(self):
        self.ws.add_to_team(
            user=self.user1.getId()
        )
        self.ws[self.user1.getId()].update({'groups': set([u'Admins'])})
        self.assertIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Members:' + self.workspace.UID()),
        )
        self.assertIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Admins:' + self.workspace.UID()),
        )

    def test_direct_set_of_membership_property_is_blocked(self):
        self.ws.add_to_team(
            user=self.user1.getId()
        )
        try:
            self.ws[self.user1.getId()].position = u'Tester'
        except Exception as e:
            self.assertEqual(
                str(e),
                'Setting membership properties directly is not supported. '
                'Use the `update` method instead.'
            )
        else:
            self.fail('Expected exception')

    def test_local_role_team_member(self):
        self.ws.add_to_team(
            user=self.user1.getId()
        )
        self.assertIn('TeamMember', self._get_roles_in_workspace())

    def test_remove_from_team(self):
        self.ws.add_to_team(
            user=self.user1.getId()
        )
        self.ws.remove_from_team(
            user=self.user1.getId()
        )
        self.assertNotIn(self.user1.getId(), list(self.ws.members))

    def test_removing_team_member_updates_groups(self):
        self.ws.add_to_team(
            user=self.user1.getId(),
            groups=(u'Admins',),
        )
        self.ws.remove_from_team(
            user=self.user1.getId()
        )
        self.assertNotIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Members:' + self.workspace.UID()),
        )
        self.assertNotIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Admins:' + self.workspace.UID()),
        )

    def test_reparent_team_member(self):
        self.ws.add_to_team(
            user=self.user1.getId(),
            groups=(u'Admins',),
            )
        user2 = api.user.create(
            email='user2@example.com',
            username='user2',
            password='123'
        )
        self.ws[self.user1.getId()].update({'user': user2.getId()})
        self.assertNotIn(self.user1.getId(), self.workspace._team)
        self.assertIn(user2.getId(), self.workspace._team)
        self.assertEqual(self.ws[user2.getId()].user, user2.getId())
        self.assertNotIn(
            self.user1.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Admins:' + self.workspace.UID()),
        )
        self.assertIn(
            user2.getId(),
            self.portal.portal_groups.getGroupMembers(
                'Admins:' + self.workspace.UID()),
        )

    def test_removing_user_removes_workspace_memberships(self):
        userid = self.user1.getId()
        self.ws.add_to_team(user=userid)
        self.assertIn(userid, self.ws.members)
        api.user.delete(userid)
        self.assertNotIn(userid, self.ws.members)
        self.assertNotIn(
            userid,
            self.portal.portal_groups.getGroupMembers(
                'Members:' + self.workspace.UID()),
        )

    def test_copying_workspace_clears_roster(self):
        cookie = self.portal.manage_copyObjects(ids=('a-workspace',))
        self.portal.manage_pasteObjects(cookie)

        self.assertTrue('copy_of_a-workspace' in self.portal)
        workspace = self.portal['copy_of_a-workspace']
        self.assertEqual(len(workspace._team), 0)
        self.assertEqual(workspace._counters['members'](), 0)

    def test_add_guest_to_team(self):
        self.ws.add_to_team(
            user=self.user1.getId(), groups=['Guests']
        )
        self.assertIn(self.user1.getId(), list(self.ws.members))

    def test_guest_has_no_team_member_role(self):
        self.ws.add_to_team(
            user=self.user1.getId(), groups=['Guests']
        )
        pmt = api.portal.get_tool('portal_membership')
        member = pmt.getMemberById(self.user1.getId())
        roles = member.getRolesInContext(self.workspace)
        self.assertIn('TeamGuest', roles)
        self.assertNotIn('TeamMember', roles)

    def test_creating_and_removing_plone_groups_is_possible(self):
        test_group_id = 'My workspace testers'
        self.assertIsNone(api.group.get(test_group_id))
        api.group.create(test_group_id)
        group = api.group.get(test_group_id)
        self.assertEquals(group.getId(), test_group_id)
        api.group.delete(test_group_id)
        self.assertIsNone(api.group.get(test_group_id))

    def test_update_membership_groups(self):
        ''' Test that calling _update_groups will always add
        the Members group, unless the user is also in the Guests group
        '''
        self.assertNotIn('TeamMember', self._get_roles_in_workspace())

        # Adding the user to the workspace will grant the TeamMember role
        membership = self.ws.add_to_team(user=self.user1.getId())
        self.assertIn('TeamMember', self._get_roles_in_workspace())

        # groups should be untouched
        old = set()
        new = set()
        membership._update_groups(old, new)
        self.assertIn('TeamMember', self._get_roles_in_workspace())

        # groups should be untouched because user is already a member
        old = set()
        new = set(['Members'])
        membership._update_groups(old, new)
        self.assertIn('TeamMember', self._get_roles_in_workspace())

        # Removing the automatic group will not actually remove it
        # because it is an automatic group
        old = set(['Members'])
        new = set()
        membership._update_groups(old, new)
        self.assertIn('TeamMember', self._get_roles_in_workspace())

        # Removing the automatic group will have no effect
        old = set(['Members'])
        new = set()
        membership._update_groups(old, new)
        self.assertIn('TeamMember', self._get_roles_in_workspace())

        # Adding the user to the Guests group will remove it from the Members
        # because the condition will not be satisifed
        old = set()
        new = set(['Guests'])
        membership._update_groups(old, new)
        self.assertIn('TeamGuest', self._get_roles_in_workspace())
        self.assertNotIn('TeamMember', self._get_roles_in_workspace())

    def test_membership_remove_all_groups(self):
        ''' Test that calling _remove_all_groups does so
        '''
        self.assertNotIn('TeamMember', self._get_roles_in_workspace())

        # Adding the user to the workspace will grant the TeamMember role
        membership = self.ws.add_to_team(user=self.user1.getId())
        self.assertIn('TeamMember', self._get_roles_in_workspace())

        # groups should be removed
        membership._remove_all_groups()
        self.assertNotIn('TeamMember', self._get_roles_in_workspace())



from Acquisition import aq_chain
from collective.workspace.interfaces import _
from collective.workspace.interfaces import IWorkspace
from plone import api
from z3c.formwidget.query.interfaces import IQuerySource
from zope.interface import classProvides
from zope.interface import directlyProvides
from zope.interface import implements
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


def find_workspace(context):
    while hasattr(context, 'context'):
        context = context.context
    for context in aq_chain(context):
        workspace = IWorkspace(context, None)
        if workspace is not None:
            return workspace


def TeamGroupsVocabulary(context):
    workspace = find_workspace(context)
    # Membership in the auto_groups is determined automatically,
    # so we don't need to show them as explicit options.
    groups = (
        set(workspace.available_groups.keys()) -
        set(workspace.auto_groups.keys())
    )
    items = []
    for group in groups:
        items.append(SimpleTerm(group, group, _(group)))
    return SimpleVocabulary(items)


directlyProvides(TeamGroupsVocabulary, IVocabularyFactory)


class UsersSource(object):
    """A source for looking up users.

    Unfortunately the one in plone.app.vocabularies is not
    quite workable with z3c.formwidget.query
    """
    implements(IQuerySource)
    classProvides(IContextSourceBinder)

    def __init__(self, context):
        self._context = context
        self._users = api.portal.get_tool('acl_users')

    def __contains__(self, value):
        return self._users.getUserById(value, None) and True or False

    def search(self, query):
        users = set()
        for u in self._users.searchUsers(id=query):
            users.add(u['userid'])
        for u in self._users.searchUsers(fullname=query):
            users.add(u['userid'])
        for u in users:
            yield self.getTerm(u)

    def getTerm(self, userid):
        fullname = userid
        user = self._users.getUserById(userid, None)
        if user:
            fullname = user.getProperty('fullname', None) or userid
        return SimpleTerm(userid, userid, fullname)
    getTermByToken = getTerm

    def __iter__(self):
        for item in self._users.searchUsers():
            yield self.getTerm(item['userid'])

    def __len__(self):
        return len(self._users.searchUsers())



from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from borg.localrole.interfaces import ILocalRoleProvider
from collective.workspace.interfaces import _
from collective.workspace.interfaces import IWorkspace
from plone import api
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PlonePAS.plugins.group import GroupManager
from zope.interface import implements


PLUGIN_ID = 'workspace_groups'


manage_addWorkspaceGroupManagerForm = PageTemplateFile(
    'templates/WorkspaceGroupManagerForm',
    globals(),
    __name__='manage_addWorkspaceGroupManagerForm'
)


def addWorkspaceGroupManager(dispatcher, id, title=None, REQUEST=None):
    """ Add a WorkspaceGroupManager to a Pluggable Auth Service. """

    pmm = WorkspaceGroupManager(id, title)
    dispatcher._setObject(pmm.getId(), pmm)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
            '%s/manage_workspace?manage_tabs_message='
            'WorkspaceGroupManager+added.'
            % dispatcher.absolute_url()
        )


class WorkspaceGroupManager(GroupManager):
    """PAS plugin to store groups created from the team rosters.

    We use a separate plugin from the standard Plone group plugin
    so that enumeration can optionally be disabled to prevent
    workspace groups from showing up in sharing and the groups
    control panel.
    """

    meta_type = 'collective.workspace Group Manager'

    security = ClassSecurityInfo()

    def removeGroup(self, group_id):
        # Don't break when PAS asks to delete a non-workspace group.
        if self.getGroupInfo(group_id) is None:
            return
        return super(WorkspaceGroupManager, self).removeGroup(group_id)

    def allowGroupAdd(self, user_id, group_id):
        # Disable adding to workspace groups in control panel
        return False

    def allowGroupRemove(self, user_id, group_id):
        # Disable removing from workspace groups in control panel.
        return False

    def allowDeletePrincipal(self, principal_id):
        # Disable deleting workspace groups in control panel.
        return False


InitializeClass(WorkspaceGroupManager)


def get_workspace_groups_plugin():
    acl_users = api.portal.get_tool('acl_users')
    return getattr(acl_users, PLUGIN_ID)


def add_group(group_id, title):
    workspace_groups = get_workspace_groups_plugin()
    if group_id not in workspace_groups._groups:
        workspace_groups.addGroup(group_id)
    gtool = api.portal.get_tool('portal_groups')
    group = gtool.getGroupById(group_id)
    group.setGroupProperties({'title': title})


class WorkspaceRoles(object):
    """Automatically assign local roles to workspace groups.
    """
    implements(ILocalRoleProvider)

    def __init__(self, context):
        self.workspace = IWorkspace(context)
        self.uid = context.UID()

    def getAllRoles(self):
        for group_name, roles in self.workspace.available_groups.items():
            group_id = group_name.encode('utf8') + ':' + self.uid
            yield group_id, roles

    def getRoles(self, user_id):
        for group_id, group_roles in self.getAllRoles():
            if user_id == group_id:
                return group_roles
        return ()


# Make the MemberAdmin role show up on the Sharing tab
class TeamManagerRoleDelegation(object):
    title = _(u"Can edit roster")
    required_permission = "collective.workspace: Manage roster"



from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting
from plone.app.robotframework.testing import AUTOLOGIN_LIBRARY_FIXTURE
from plone.testing import z2

from zope.configuration import xmlconfig


class CollectiveWorkspaceLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load ZCML
        import collective.workspace
        xmlconfig.file(
            'configure.zcml',
            collective.workspace,
            context=configurationContext
        )

        z2.installProduct(app, 'collective.workspace')

    def tearDownZope(self, app):
        # Uninstall products installed above
        z2.uninstallProduct(app, 'collective.workspace')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'collective.workspace:default')

        # Create a content type with the behavior enabled
        from plone.dexterity.fti import DexterityFTI
        fti = DexterityFTI('Workspace')
        fti.behaviors = (
            'plone.app.content.interfaces.INameFromTitle',
            'plone.app.dexterity.behaviors.metadata.IBasic',
            'collective.workspace.interfaces.IWorkspace',
            )
        portal.portal_types._setObject('Workspace', fti)


COLLECTIVE_WORKSPACE_FIXTURE = CollectiveWorkspaceLayer()
COLLECTIVE_WORKSPACE_INTEGRATION_TESTING = IntegrationTesting(
    bases=(COLLECTIVE_WORKSPACE_FIXTURE,),
    name="CollectiveWorkspaceLayer:Integration"
)
COLLECTIVE_WORKSPACE_ROBOT_TESTING = FunctionalTesting(
    bases=(
        COLLECTIVE_WORKSPACE_FIXTURE, AUTOLOGIN_LIBRARY_FIXTURE, z2.ZSERVER),
    name="CollectiveWorkspaceLayer:Robot"
)



from zope.i18nmessageid import MessageFactory
from zope.interface import Interface

_ = MessageFactory('collective.workspace')


class IHasWorkspace(Interface):
    """Marker for team workspaces."""


class IWorkspace(Interface):
    """Interface for interacting with an item's workspace features."""


class IRosterView(Interface):
    """Marker for display view of a roster."""



from BTrees.Length import Length
from collective.workspace.events import TeamMemberModifiedEvent
from collective.workspace.events import TeamMemberRemovedEvent
from collective.workspace.interfaces import _
from collective.workspace.interfaces import IWorkspace
from collective.workspace.pas import add_group
from collective.workspace.pas import get_workspace_groups_plugin
from collective.workspace.vocabs import UsersSource
from copy import deepcopy
from DateTime import DateTime
from plone import api
from plone.autoform import directives as form
from plone.formwidget.autocomplete import AutocompleteFieldWidget
from plone.supermodel import model
from plone.uuid.interfaces import IUUIDGenerator
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope import schema
from zope.component import adapter
from zope.component import getUtility
from zope.event import notify
from zope.interface import implementer


class ITeamMembership(model.Schema):
    """Schema for one person's membership in a team."""

    form.widget(user=AutocompleteFieldWidget)
    user = schema.Choice(
        title=_(u'User'),
        source=UsersSource,
    )

    position = schema.TextLine(
        title=_(u'Position'),
        required=False,
    )

    form.widget(groups=CheckBoxFieldWidget)
    groups = schema.Set(
        title=_(u'Groups'),
        required=False,
        value_type=schema.Choice(
            vocabulary='collective.workspace.groups',
        ),
    )


@implementer(ITeamMembership)
class TeamMembership(object):

    __slots__ = ('workspace', '__dict__')
    _schema = ITeamMembership

    def __init__(self, workspace, data):
        self.workspace = workspace
        if 'UID' not in data:
            data['UID'] = getUtility(IUUIDGenerator)()
        self.__dict__ = data

    @property
    def _key(self):
        return self.user or self.UID

    @property
    def _title(self):
        mtool = api.portal.get_tool('portal_membership')
        member = mtool.getMemberById(self._key)
        if member is not None:
            return member.getProperty('fullname') or self._key
        else:
            return self._key

    def __getattr__(self, name):
        field = self.__class__._schema.get(name, None)
        if field is None:
            raise AttributeError(name)
        field = field.bind(self)
        return deepcopy(field.default)

    def __setattr__(self, name, value):
        if name not in ('workspace', '__dict__'):
            raise Exception(
                'Setting membership properties directly is not supported. '
                'Use the `update` method instead.'
            )
        super(TeamMembership, self).__setattr__(name, value)

    def _update_groups(self, old_groups, new_groups):
        workspace = self.workspace
        context = workspace.context
        uid = context.UID()
        workspace_groups = get_workspace_groups_plugin()
        if self.user is None:
            return

        for group_name in new_groups:
            if group_name not in workspace.available_groups:
                raise ValueError(
                    'Unknown workspace group: {}'.format(group_name))
        if self.groups != new_groups:
            self.__dict__['groups'] = new_groups.copy()

        # Determine automatic groups
        new_groups = set(new_groups)
        old_groups = set(old_groups)
        for name, condition in workspace.auto_groups.items():
            if name not in workspace.available_groups:
                raise Exception('Unknown workspace group: {}'.format(name))
            # only add the automatic groups if condition is satisfied,
            # otherwise remove it
            if condition(self):
                new_groups.add(name)
            else:
                old_groups.add(name)

        # Add to new groups
        for group_name in (new_groups - old_groups):
            group_id = '{}:{}'.format(group_name.encode('utf8'), uid)
            try:
                workspace_groups.addPrincipalToGroup(self.user, group_id)
            except KeyError:  # group doesn't exist
                title = '{}: {}'.format(
                    group_name.encode('utf8'), context.Title())
                add_group(group_id, title)
                workspace_groups.addPrincipalToGroup(self.user, group_id)

        # Remove from old groups
        for group_name in (old_groups - new_groups):
            group_id = '{}:{}'.format(group_name.encode('utf8'), uid)
            try:
                workspace_groups.removePrincipalFromGroup(self.user, group_id)
            except KeyError:  # group doesn't exist
                pass

    def _remove_all_groups(self):
        workspace = self.workspace
        uid = workspace.context.UID()
        workspace_groups = get_workspace_groups_plugin()
        groups = set(self.groups) | set(workspace.available_groups)
        for group_name in groups:
            group_id = '{}:{}'.format(group_name.encode('utf8'), uid)
            try:
                workspace_groups.removePrincipalFromGroup(self.user, group_id)
            except KeyError:  # group doesn't exist
                pass

    @property
    def groups(self):
        # Don't include automatic groups
        groups = set(self.__dict__.get('groups') or set())
        groups -= set(self.workspace.auto_groups.keys())
        return groups

    def update(self, data):
        old = self.__dict__.copy()
        old_key = self._key
        user_changed = False
        if 'user' in data and old['user'] != data['user']:
            # User is changing, so remove the old user from groups.
            user_changed = True
            self._remove_all_groups()
        data['_mtime'] = DateTime()
        self.__dict__.update(data)
        # make sure change is persisted
        # XXX really we should use PersistentDicts
        workspace = self.workspace
        if user_changed:
            # User changed; remove old entry in _team
            del workspace.context._team[old_key]
            # Add new user to groups
            self._update_groups(set(), self.groups)
        workspace.context._team[self._key] = self.__dict__

        # update counters
        for name, func in workspace.counters:
            # The following is based on Python's ability to treat booleans
            # as integers (1 or 0)...
            # 0 = no change
            # 1 = matches now but didn't before
            # -1 = matched before but doesn't now
            # Then we use that difference to update the count
            # of how many roster members match.
            diff = func(self.__dict__) - func(old)
            if diff:
                if name not in workspace.context._counters:
                    workspace.context._counters[name] = Length()
                workspace.context._counters[name].change(diff)

        if 'groups' in data:
            self._update_groups(old['groups'], data['groups'])

        self.handle_modified(old)
        notify(TeamMemberModifiedEvent(self.workspace.context, self))
        if user_changed:
            workspace.context.reindexObject(idxs=['workspace_members'])

    def handle_added(self):
        pass

    def handle_modified(self, old):
        pass

    def handle_removed(self):
        pass

    def remove_from_team(self):
        workspace = self.workspace
        for name, func in workspace.counters:
            if func(self.__dict__):
                workspace.context._counters[name].change(-1)
        del self.workspace.members[self._key]
        self._remove_all_groups()
        self.handle_removed()
        notify(TeamMemberRemovedEvent(self.workspace.context, self))
        self.workspace.context.reindexObject(idxs=['workspace_members'])


@adapter(ITeamMembership)
@implementer(IWorkspace)
def workspace_from_membership(membership):
    return membership.workspace



from .interfaces import IHasWorkspace
from .interfaces import IWorkspace
from plone.indexer import indexer
from zope.interface import Interface


@indexer(IHasWorkspace)
def workspace_members(self):
    workspace = IWorkspace(self)
    return set(workspace.members)


@indexer(Interface)
def null_indexer(self):
    raise AttributeError



from collective.workspace.testing import COLLECTIVE_WORKSPACE_ROBOT_TESTING
from plone.testing import layered
import robotsuite
import unittest


def test_suite():
    suite = unittest.TestSuite()
    suite.addTests([
        layered(robotsuite.RobotTestSuite("workspace.robot"),
                layer=COLLECTIVE_WORKSPACE_ROBOT_TESTING)
    ])
    return suite



# coding=utf-8
from zope.component.interfaces import ObjectEvent


class TeamMemberEvent(ObjectEvent):
    """Base for team member events.

    This is used instead of normal Zope object lifecycle
    events so it can be dispatched based on the workspace
    rather than based on the modified object (which is
    just a dictionary in this case).
    """

    def __init__(self, object, membership):
        self.object = object
        self.membership = membership


class TeamMemberAddedEvent(TeamMemberEvent):
    """Event for when a member is added."""


class TeamMemberRemovedEvent(TeamMemberEvent):
    """Event for when a member is removed."""


class TeamMemberModifiedEvent(TeamMemberEvent):
    """Event for when a member is modified."""



from .events import TeamMemberAddedEvent
from .interfaces import IHasWorkspace
from .interfaces import IWorkspace
from .membership import ITeamMembership
from .membership import TeamMembership
from .pas import add_group
from .pas import get_workspace_groups_plugin
from BTrees.Length import Length
from BTrees.OOBTree import OOBTree
from DateTime import DateTime
from plone import api
from plone.uuid.interfaces import IUUIDGenerator
from Products.PluggableAuthService.interfaces.events import IPrincipalDeletedEvent  # noqa: E501
from zope.component import adapter
from zope.component import getUtility
from zope.container.interfaces import IObjectAddedEvent
from zope.container.interfaces import IObjectRemovedEvent
from zope.event import notify
from zope.lifecycleevent.interfaces import IObjectCopiedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent


class Workspace(object):
    """Provides access to team workspace functionality.

    This is registered as a behavior providing the IWorkspace interface.
    """

    def __init__(self, context):
        self.context = context

        # create a BTree to store team member data
        if not hasattr(self.context, '_team'):
            self.context._team = OOBTree()

        if not hasattr(self.context, '_counters'):
            self._recount()

    def _recount(self):
        counters = {}
        for name, func in self.counters:
            counters[name] = Length()
        for m in self.context._team.itervalues():
            for name, func in self.counters:
                if func(m):
                    counters[name].change(1)
        self.context._counters = counters

    @property
    def membership_schema(self):
        """Returns the schema to be used for editing team memberships.
        """
        return ITeamMembership

    membership_factory = TeamMembership

    # A list of groups to which team members can be assigned.
    # Maps group name -> roles
    available_groups = {
        u'Members': ('Contributor', 'Reader', 'TeamMember'),
        u'Guests': ('TeamGuest', ),
        u'Admins': ('Contributor', 'Editor', 'Reviewer',
                    'Reader', 'TeamManager',),
    }

    # Add everyone on the roster to the Members group
    auto_groups = {
        u'Members': lambda x: 'Guests' not in set(x.groups),
    }

    counters = (
        ('members', lambda x: True),
    )

    @property
    def members(self):
        """Access the raw team member data."""
        return self.context._team

    def __getitem__(self, name):
        memberdata = self.context._team[name]
        return self.membership_factory(self, memberdata)

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def __iter__(self):
        for key in self.context._team.iterkeys():
            yield self[key]

    def add_to_team(self, user, groups=None, **kw):
        """
        Makes sure a user is in this workspace's team.

        :param user: The id of the user to add to this workspace
        :type user: str
        :param groups: The set of workspace groups to add this user to
        :type groups: set
        :param kw: Pass user and any other attributes that should be set on the
                   team member.
        :type kw: dict
        """
        # TODO: user argument should be renamed to userid for clarity
        #       however doing so now would break backwards compatibility
        data = kw.copy()
        data['user'] = user
        if groups is not None:
            data['groups'] = groups = set(groups)
        members = self.members
        if not user or user not in members:
            data['UID'] = uid = getUtility(IUUIDGenerator)()
            key = user or uid
            data['_mtime'] = DateTime()
            if groups is None:
                data['groups'] = groups = set()
            members[key] = data
            for name, func in self.counters:
                if func(data):
                    if name not in self.context._counters:
                        self.context._counters[name] = Length()
                    self.context._counters[name].change(1)
            membership = self.membership_factory(self, data)
            membership.handle_added()
            membership._update_groups(set(), groups)
            notify(TeamMemberAddedEvent(self.context, membership))
            self.context.reindexObject(
                idxs=['workspace_members', 'workspace_leaders']
            )
        else:
            membership = self.membership_factory(self, self.members[user])
            membership.update(data)
        return membership

    def remove_from_team(self, user):
        """
        Remove a user from the workspace

        :param user: The id of the user to remove from this workspace
        :type user: str
        """
        # TODO: user argument should be renamed to key for clarity
        #       however doing so now would break backwards compatibility
        key = user
        membership = self.get(key)
        if membership is not None:
            membership.remove_from_team()
        return membership


@adapter(IHasWorkspace, IObjectAddedEvent)
def handle_workspace_added(context, event):
    workspace = IWorkspace(context)
    for group_name in workspace.available_groups:
        group_id = '{}:{}'.format(group_name.encode('utf8'), context.UID())
        title = '{}: {}'.format(group_name.encode('utf8'), context.Title())
        add_group(group_id, title)


@adapter(IHasWorkspace, IObjectModifiedEvent)
def handle_workspace_modified(context, event):
    workspace = IWorkspace(context)
    gtool = api.portal.get_tool('portal_groups')
    for group_name in workspace.available_groups:
        group_id = '{}:{}'.format(group_name.encode('utf8'), context.UID())
        group_title = '{}: {}'.format(
            group_name.encode('utf8'), context.Title())
        group = gtool.getGroupById(group_id)
        if group is not None:
            group.setProperties(title=group_title)


@adapter(IHasWorkspace, IObjectRemovedEvent)
def handle_workspace_removed(context, event):
    workspace = IWorkspace(context)
    workspace_groups = get_workspace_groups_plugin()
    for group_name in workspace.available_groups:
        group_id = '{}:{}'.format(group_name.encode('utf8'), context.UID())
        try:
            workspace_groups.removeGroup(group_id)
        except KeyError:  # group already doesn't exist
            pass


@adapter(IHasWorkspace, IObjectCopiedEvent)
def handle_workspace_copied(context, event):
    if hasattr(context, '_team'):
        del context._team
    if hasattr(context, '_counters'):
        del context._counters


@adapter(IPrincipalDeletedEvent)
def handle_principal_deleted(event):
    """When a user is deleted, remove it from all workspaces."""
    principal = event.principal
    catalog = api.portal.get_tool('portal_catalog')
    for b in catalog.unrestrictedSearchResults(workspace_members=principal):
        workspace = IWorkspace(b._unrestrictedGetObject(), None)
        if workspace is not None:
            membership = workspace.get(principal)
            if membership is not None:
                membership.remove_from_team()



from setuptools import setup, find_packages

version = '2.0b3'

long_description = (
    open('README.rst').read()
    + '\n' +
    'Contributors\n'
    '============\n'
    + '\n' +
    open('CONTRIBUTORS.rst').read()
    + '\n' +
    open('CHANGES.rst').read()
    + '\n')

setup(name='collective.workspace',
      version=version,
      description="Provide 'membership' in specific areas of a Plone Site",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          "Environment :: Web Environment",
          "Framework :: Plone",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Topic :: Software Development :: Libraries :: Python Modules",
      ],
      keywords='plone workspace collaboration',
      author='David Glick',
      author_email='',
      url='https://github.com/collective/collective.workspace',
      license='gpl',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['collective', ],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'plone.api',
          'plone.app.dexterity',
          'plone.formwidget.autocomplete',
          # -*- Extra requirements: -*-
      ],
      extras_require={
          'test': [
              'plone.app.robotframework',
              'plone.app.testing',
          ],
          'develop': [
              'zest.releaser',
          ],
      },
      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
)



from collective.workspace.testing import \
    COLLECTIVE_WORKSPACE_INTEGRATION_TESTING
import unittest2 as unittest


class TestExample(unittest.TestCase):

    layer = COLLECTIVE_WORKSPACE_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']
        self.qi_tool = self.portal.portal_quickinstaller

    def test_product_is_installed(self):
        """ Validate that our products GS profile has been run and the product
            installed
        """
        pid = 'collective.workspace'
        installed = [p['id'] for p in self.qi_tool.listInstalledProducts()]
        self.assertTrue(pid in installed,
                        'package appears not to have been installed')



from AccessControl.Permissions import add_user_folders
from collective.workspace import pas
from Products.PluggableAuthService import registerMultiPlugin


registerMultiPlugin(pas.WorkspaceGroupManager.meta_type)


def initialize(context):

    context.registerClass(
        pas.WorkspaceGroupManager,
        permission=add_user_folders,
        constructors=(pas.manage_addWorkspaceGroupManagerForm,
                      pas.addWorkspaceGroupManager),
        visibility=None,
    )



from AccessControl import getSecurityManager
from Products.statusmessages.interfaces import IStatusMessage
from collective.workspace.interfaces import _
from collective.workspace.interfaces import IRosterView
from collective.workspace.interfaces import IWorkspace
from collections import namedtuple
from plone.autoform.base import AutoFields
from plone.autoform.form import AutoExtensibleForm
from plone.z3cform import z2
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from z3c.form.form import DisplayForm
from z3c.form.form import EditForm
from z3c.form.interfaces import ActionExecutionError
from zope.cachedescriptors.property import Lazy as lazy_property
from zope.interface import implementer
from zope.publisher.interfaces.browser import IPublishTraverse
import transaction


Action = namedtuple('Action', "label view_name permission")


@implementer(IRosterView)
class TeamRosterView(AutoFields, DisplayForm):
    """Display the roster as a table."""

    row_template = ViewPageTemplateFile('templates/team_roster_row.pt')

    row_actions = (
        Action('Edit', 'edit-roster', 'collective.workspace: Manage roster'),
    )

    table_actions = ()

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.workspace = IWorkspace(self.context)

    @property
    def label(self):
        return _(u'Roster: ${title}', mapping={'title': self.context.Title()})

    @lazy_property
    def schema(self):
        return self.workspace.membership_schema

    ignoreContext = True

    def getContent(self):
        return self.context

    def row_views(self):
        for membership in self.workspace:
            for widget in self.widgets.values():
                widget.ignoreContext = False
                widget.context = membership
                widget.update()
            yield self.row_template(membership=membership)

    @lazy_property
    def can_edit_roster(self):
        return getSecurityManager().checkPermission(
            'collective.workspace: Manage roster', self.context)

    def update(self):
        z2.switch_on(self)
        self.updateFieldsFromSchemata()
        self.updateWidgets()

    def render(self):
        return self.index()

    def __call__(self):
        self.update()
        return self.render()


@implementer(IPublishTraverse)
class TeamMemberEditForm(AutoExtensibleForm, EditForm):

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.workspace = IWorkspace(self.context)

    key = None

    def publishTraverse(self, request, name):
        self.key = name
        return self

    @lazy_property
    def schema(self):
        return self.workspace.membership_schema

    def updateFields(self):
        super(TeamMemberEditForm, self).updateFields()
        # don't show the user field if we are editing
        if self.key:
            del self.fields['user']

    @lazy_property
    def ignoreContext(self):
        return not bool(self.key)

    @lazy_property
    def label(self):
        if self.key:
            return self.getContent()._title
        else:
            return _(u'Add Person to Roster')

    @lazy_property
    def _content(self):
        if not self.key:
            return self.context
        workspace = self.workspace
        memberdata = workspace.members[self.key]
        return workspace.membership_factory(workspace, memberdata)

    def getContent(self):
        return self._content

    def validateInvariants(self, membership):
        pass

    @button.buttonAndHandler(_(u'Save'))
    def handleSave(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        status = _(u'Changes saved')

        if self.key:
            membership = self.getContent()
            membership.update(data)
        else:
            # Add new roster member
            membership = self.workspace.add_to_team(**data)
            status = _(u'User added')

        try:
            self.validateInvariants(membership)
        except ActionExecutionError:
            # make sure above changes won't be persisted
            transaction.abort()
            raise

        self._finished = True
        IStatusMessage(self.request).addStatusMessage(status, 'info')

    @property
    def can_remove(self):
        return self.key

    @button.buttonAndHandler(
        _(u'Remove'), condition=lambda self: self.can_remove)
    def handleRemove(self, action):
        membership = self.getContent()
        membership.remove_from_team()
        self._finished = True
        IStatusMessage(self.request).addStatusMessage(
            _(u"User removed"), "info")

    _finished = False

    def render(self):
        if self._finished:
            return ' '
        return super(TeamMemberEditForm, self).render()
